/* $Id: ipfilter.l,v 1.13 2009/10/12 19:31:10 bhockney Exp $ */

/* (C) 2004-2009 by Bob Hockney <zeus@ix.netcom.com>    *
 * Based on fwlogwatch written by                       *
 * Boris Wesslowski <Wesslowski@CERT.Uni-Stuttgart.DE>  *
 *                                                      *
 * ipfilter log parser for wfwl_syslog                  *
 *                                                      *
 * This code is distributed under the terms of GNU GPL  */

%option prefix="ipf"
%option outfile="ipfilter.c"
%option noyywrap
%option nounput
%option noinput

%{
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>
#include "main.h"
#include "utils.h"

#ifndef IP_DF
#define IP_DF 0x4000
#endif
#ifndef IP_MF
#define IP_MF 0x2000
#endif

extern struct options opt;

void ipf_parse_date(char *input);
void ipf_parse_data(char *input, unsigned char mode);
void ipf_parse_ips(char *input, unsigned char mode);
void ipf_parse_proto(char *input);
void ipf_parse_frag(char *input);
void ipf_parse_flags(char *input);
void ipf_parse_seq(char *input);
%}

MONTH		"Jan"|"Feb"|"Mar"|"Apr"|"May"|"Jun"|"Jul"|"Aug"|"Sep"|"Oct"|"Nov"|"Dec"
STRING		[a-zA-Z0-9.,_-]*
LOGHOST		[0-9.a-zA-Z()_:-]*
DIGIT		[0-9]
NUMBER		{DIGIT}+
HEXDIGIT	[0-9a-f]
OCTET		{DIGIT}{1,3}
PORT		{DIGIT}{1,5}
IPFILTER	"ipmon["{NUMBER}"]:"
FLAGS		"-"[SAFRPU]+
PROTO		[0-9a-z-]+
TARGET		[SpPbBnL]
FRAG        "frag"[ ][+-]{0,2}{NUMBER}"@"{NUMBER} 

%%

{MONTH}[ ]{1,2}{DIGIT}{1,2}[ ]{DIGIT}{2}:{DIGIT}{2}:{DIGIT}{2}[ ]{LOGHOST}	ipf_parse_date(ipftext);
{IPFILTER}														/* ignore */
"[ID "{NUMBER}" "{STRING}"]"									/* ignore */
{DIGIT}{2}":"{DIGIT}{2}":"{DIGIT}{2}"."{DIGIT}{6}				/* ignore */
{STRING}[ ]"@"{NUMBER}":"{NUMBER}[ ]{TARGET}					ipf_parse_data(ipftext, IPF_OPT_NONE);
{NUMBER}"x"[ ]+{STRING}[ ]"@"{NUMBER}":"{NUMBER}[ ]{TARGET}		ipf_parse_data(ipftext, IPF_OPT_COUNT);
{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}							ipf_parse_ips(ipftext, IPF_OPT_SRC);
" -> "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}						ipf_parse_ips(ipftext+4, IPF_OPT_DST);
{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{PORT}					ipf_parse_ips(ipftext, IPF_OPT_SRC|IPF_OPT_PORT);
" -> "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{PORT}			ipf_parse_ips(ipftext+4, IPF_OPT_DST|IPF_OPT_PORT);
{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"]"				ipf_parse_ips(ipftext, IPF_OPT_SRC|IPF_OPT_RES);
" -> "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"]"		ipf_parse_ips(ipftext+4, IPF_OPT_DST|IPF_OPT_RES);
{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}		ipf_parse_ips(ipftext, IPF_OPT_SRC|IPF_OPT_RES|IPF_OPT_PORT);
" -> "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}	ipf_parse_ips(ipftext+4, IPF_OPT_DST|IPF_OPT_RES|IPF_OPT_PORT);
{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{STRING}				ipf_parse_ips(ipftext, IPF_OPT_SRC|IPF_OPT_RPORT);
" -> "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{STRING}			ipf_parse_ips(ipftext+4, IPF_OPT_DST|IPF_OPT_RPORT);
{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{STRING}	ipf_parse_ips(ipftext, IPF_OPT_SRC|IPF_OPT_RES|IPF_OPT_RPORT);
" -> "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{STRING}		ipf_parse_ips(ipftext+4, IPF_OPT_DST|IPF_OPT_RES|IPF_OPT_RPORT);
"PR "{PROTO}									ipf_parse_proto(ipftext+3);
"len "{NUMBER}[ ][(]?{NUMBER}[)]?				sscanf(ipftext+4, "%d%*2[ (]%d", &opt.line->ihl, &opt.line->datalen);
{FRAG}											ipf_parse_frag(ipftext+6);
"icmp "{DIGIT}{1,2}"/"{DIGIT}{1,2}				sscanf(ipftext, "icmp %d/%d", &opt.line->icmp_type, &opt.line->icmp_code); opt.parser=opt.parser|IPF_TYPE;
"icmp "{STRING}"/"{STRING}						if (get_icmp_values(ipftext+5) == PARSE_OK); opt.parser=opt.parser|IPF_TYPE;
"for "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{PORT}" - "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{PORT}" PR "{PROTO}" len "{NUMBER}[ ]{NUMBER}{FRAG}?	/* ignore */
"for "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{STRING}" - "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{STRING}" PR "{PROTO}" len "{NUMBER}[ ]{NUMBER}{FRAG}?	/* ignore */
"for "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}" - "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{PORT}" PR "{PROTO}" len "{NUMBER}[ ]{NUMBER}{FRAG}?	/* ignore */
"for "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}" - "{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}","{STRING}" PR "{PROTO}" len "{NUMBER}[ ]{NUMBER}{FRAG}?	/* ignore */
"for "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"] - "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"] PR "{PROTO}" len "{NUMBER}" ("{NUMBER}")"{FRAG}?	/* ignore */
"for "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}" - "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}" PR "{PROTO}" len "{NUMBER}[ ]{NUMBER}{FRAG}?	/* ignore */
"for "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{PORT}" - "{STRING}"["{OCTET}"."{OCTET}"."{OCTET}"."{OCTET}"],"{STRING}" PR "{PROTO}" len "{NUMBER}[ ]{NUMBER}{FRAG}?	/* ignore */
{FLAGS}											ipf_parse_flags(ipftext+1);
{NUMBER}[ ]{NUMBER}[ ]{NUMBER}					ipf_parse_seq(ipftext);
"K-S"											/* ignore */
"K-F"											/* ignore */
"IN"											strcpy(opt.line->inif, opt.line->interface);
"OUT"											strcpy(opt.line->outif, opt.line->interface);
"mbcast"	/* ignore */
({HEXDIGIT}{HEXDIGIT}[ ]?)+[ ]+[ -~]+			opt.parser=opt.parser|IPF_NO_HIT;
[ \t]+		/* ignore whitespace */
[\n]		/* ignore */
{STRING}	if(opt.verbose) fprintf(stderr, "Unrecognized token: %s\n", ipftext);
.		if(opt.verbose) fprintf(stderr, "Unrecognized character: %s\n", ipftext);

%%

void ipf_parse_date(char *input)
{
  int retval, day, hour, minute, second;
  char smonth[4];

  retval = sscanf(input, "%3s %2d %2d:%2d:%2d %32s",
		  smonth, &day, &hour, &minute, &second,
		  opt.line->hostname);
  if (retval != 6) return;

  build_time(smonth, day, hour, minute, second);

  opt.parser=opt.parser|IPF_DATE;
}

void ipf_parse_data(char *input, unsigned char mode)
{
  int retval;

  if (mode == IPF_OPT_COUNT) {
    retval = sscanf(input, "%lux %" SHORTLEN_S "s @%" SHORTLEN_S "s %" SHORTLEN_S "s",
		    &opt.line->count,
		    opt.line->interface,
		    opt.line->log_label,
		    opt.line->branchname);
    if (retval != 4) return;
  } else {
    retval = sscanf(input, "%" SHORTLEN_S "s @%" SHORTLEN_S "s %" SHORTLEN_S "s",
		    opt.line->interface,
		    opt.line->log_label,
		    opt.line->branchname);
    if (retval != 3) return;

    opt.line->count = 1;
  }

  strcat(opt.line->log_label, " ");
  strcat(opt.line->log_label, opt.line->branchname);

  opt.parser=opt.parser|IPF_DATA;
}

void ipf_parse_ips(char *input, unsigned char mode)
{
  int host1, host2, host3, host4, port;
  int retval;
  char ip[IPLEN];
  char name[SHORTLEN], *ptr, portname[SHOSTLEN];

  if (mode == (IPF_OPT_SRC|IPF_OPT_RES|IPF_OPT_RPORT) || mode == (IPF_OPT_DST|IPF_OPT_RES|IPF_OPT_RPORT)) {
    ptr = strchr(input, '['); *ptr = ' ';
    retval = sscanf(input, "%" SHORTLEN_S "s %3d.%3d.%3d.%3d],%" SHOSTLEN_S "s",
		    name, &host1, &host2, &host3, &host4, portname);
    if(retval != 6) return;
  } else if (mode == (IPF_OPT_SRC|IPF_OPT_RES|IPF_OPT_PORT) || mode == (IPF_OPT_DST|IPF_OPT_RES|IPF_OPT_PORT)) {
    ptr = strchr(input, '['); *ptr = ' ';
    retval = sscanf(input, "%" SHORTLEN_S "s %3d.%3d.%3d.%3d],%5d",
		    name, &host1, &host2, &host3, &host4, &port);
    if(retval != 6) return;
  } else if (mode == (IPF_OPT_SRC|IPF_OPT_RES) || mode == (IPF_OPT_DST|IPF_OPT_RES)) {
    ptr = strchr(input, '['); *ptr = ' ';
    retval = sscanf(input, "%" SHORTLEN_S "s %3d.%3d.%3d.%3d]",
		    name, &host1, &host2, &host3, &host4);
    if(retval != 5) return;
  } else if (mode == (IPF_OPT_SRC|IPF_OPT_RPORT) || mode == (IPF_OPT_DST|IPF_OPT_RPORT)) {
    retval = sscanf(input, "%3d.%3d.%3d.%3d,%" SHOSTLEN_S "s",
		    &host1, &host2, &host3, &host4, portname);
    if(retval != 5) return;
  } else if (mode == (IPF_OPT_SRC|IPF_OPT_PORT) || mode == (IPF_OPT_DST|IPF_OPT_PORT)) {
    retval = sscanf(input, "%3d.%3d.%3d.%3d,%5d",
		    &host1, &host2, &host3, &host4, &port);
    if(retval != 5) return;
  } else if (mode == IPF_OPT_SRC || mode == IPF_OPT_DST) {
    retval = sscanf(input, "%3d.%3d.%3d.%3d",
		    &host1, &host2, &host3, &host4);
    if(retval != 4) return;
  } else {
    return;
  }

  snprintf(ip, IPLEN, "%d.%d.%d.%d", host1, host2, host3, host4);
  if(mode & IPF_OPT_SRC) {
    if(convert_ip(ip, &opt.line->shost) == IN_ADDR_ERROR) return;
    opt.parser=opt.parser|IPF_SRC_IP;
    if(mode & IPF_OPT_RES) strcpy((char *)&opt.line->shostname, name);
    if(mode & IPF_OPT_PORT) {
      opt.line->sport = port;
      opt.parser=opt.parser|IPF_SRC_PORT;
    }
    if(mode & IPF_OPT_RPORT) strcpy((char *)&opt.line->sservice, portname);
  } else {
    if(convert_ip(ip, &opt.line->dhost) == IN_ADDR_ERROR) return;
    opt.parser=opt.parser|IPF_DST_IP;
    if(mode & IPF_OPT_RES) strcpy((char *)&opt.line->dhostname, name);
    if(mode & IPF_OPT_PORT) {
      opt.line->dport = port;
      opt.parser=opt.parser|IPF_DST_PORT;
    }
    if(mode & IPF_OPT_RPORT) strcpy((char *)&opt.line->dservice, portname);
  }
}

void ipf_parse_proto(char *input)
{
  if(isdigit((int)input[0])) {
    opt.line->protocol = atoi(input);
  } else {
    struct protoent *proto;

    proto = getprotobyname(input);
    if (proto != NULL)
      opt.line->protocol = proto->p_proto;
  }

  if (opt.line->protocol != 0) {
    opt.parser=opt.parser|IPF_PROTO;
  } else {
    fprintf(stderr, "Unknown protocol (not in /etc/protocols), ignoring: %s\n", input);
    opt.parser=opt.parser|IPF_NO_HIT;
  }
}

void ipf_parse_frag(char *input) {
  char *ptr;

  if (strchr(input, '+') != NULL) opt.line->fragoff |= IP_MF;
  if (strchr(input, '-') != NULL) opt.line->fragoff |= IP_DF;
  ptr = strchr(input, '@');
  if (ptr != NULL)
    opt.line->fragoff |= atoi(++ptr);

}

void ipf_parse_flags(char *input)
{
  while (*input != '\0') {
    if(*input == 'S') opt.line->flags |= TCP_SYN;
    if(*input == 'A') opt.line->flags |= TCP_ACK;
    if(*input == 'F') opt.line->flags |= TCP_FIN;
    if(*input == 'R') opt.line->flags |= TCP_RST;
    if(*input == 'P') opt.line->flags |= TCP_PSH;
    if(*input == 'U') opt.line->flags |= TCP_URG;
    input++;
  }
}

void ipf_parse_seq(char *input) {
  int retval;

  retval = sscanf(input, "%lu %lu %u",
    &opt.line->tcp_seq, &opt.line->tcp_ack_seq, &opt.line->tcp_window);

}

unsigned char flex_ipfilter(char *input, int linenum)
{
  struct yy_buffer_state *state;
  long int portnum = 0;

  opt.parser = 0;
  init_line();
  state = ipf_scan_string(input);
  ipflex();
  ipf_delete_buffer(state);

  if (opt.parser & IPF_NO_HIT)
    return PARSE_NO_HIT;

  if (opt.parser & IPF_PROTO && !(opt.parser & IPF_SRC_PORT)) {
    portnum = get_port((char *)&opt.line->sservice, opt.line->protocol);
    if (portnum >= 0) {
      opt.line->sport = portnum; 
      opt.parser |= IPF_SRC_PORT;
    }
  }

  if (opt.parser & IPF_PROTO && !(opt.parser & IPF_DST_PORT)) {
    portnum = get_port((char *)&opt.line->dservice, opt.line->protocol);
    if (portnum >= 0) {
      opt.line->dport = portnum; 
      opt.parser |= IPF_DST_PORT;
    }
  }

  if ((opt.line->protocol == 6 || opt.line->protocol == 17) &&
       opt.parser == (IPF_DATE|IPF_DATA|IPF_PROTO|IPF_SRC_IP|IPF_DST_IP|IPF_SRC_PORT|IPF_DST_PORT)) {
    return PARSE_OK;
  }
  if ((opt.line->protocol == 1) && opt.parser == (IPF_DATE|IPF_DATA|IPF_PROTO|IPF_SRC_IP|IPF_DST_IP|IPF_TYPE)) {
    return PARSE_OK;
  }

  if (!(opt.line->protocol == 6 || opt.line->protocol == 17 ||
        opt.line->protocol == 1) &&
       (opt.parser & (IPF_DATE|IPF_DATA|IPF_PROTO|IPF_SRC_IP|IPF_DST_IP)) ==
       (IPF_DATE|IPF_DATA|IPF_PROTO|IPF_SRC_IP|IPF_DST_IP)) {
    return PARSE_OK;
  }
  if(opt.verbose >= VERBOSE_ERROR)
    fprintf(stderr, "ipfilter parse error in line %d, ignoring.\n", linenum);
  if(opt.verbose >= VERBOSE_DEBUG)
    fprintf(stderr, "input was: \"%s\"\n", input);
  return PARSE_WRONG_FORMAT;
}
